import os, glob, shutil
import re
from Bio.Blast.Applications import NcbipsiblastCommandline
from pdb2sql import pdb2sql

from pssmgen.map_pssm2pdb import write_mapped_pssm_pdb


class PSSM(object):

    def __init__(self, work_dir='.'):

        """Compute the PSSM and map the the sequence for a series of decoys.

        Args:
            caseID (TYPE): Name of the case. This must correspond to a directory name
            pdb_dir (str, optional): directory name where the decoys pdbs are stored within caseID

        Example:

        >>> # import the module
        >>> from pssmgen import PSSM
        >>>
        >>> # create ab instance
        >>> gen = PSSM('1AK4',pdb_dir='water')
        >>>
        >>> # configure the generator
        >>> gen.configure(blast='/home/software/blast/bin/psiblast',
                         database='/data/DBs/blast_dbs/nr_v20180204/nr')
        >>>
        >>> # generates the FASTA query
        >>> gen.get_fasta()
        >>>
        >>> # generates the PSSM
        >>> gen.get_pssm()
        >>>
        >>> # map the pssm to the pdb
        >>> gen.map_pssm()
        """

        self.work_dir = work_dir

        self.One2ThreeDict = {
        'A' : 'ALA', 'R' : 'ARG', 'N' : 'ASN', 'D' : 'ASP', 'C' : 'CYS', 'E' : 'GLU', 'Q' : 'GLN',
        'G' : 'GLY', 'H' : 'HIS', 'I' : 'ILE', 'L' : 'LEU', 'K' : 'LYS', 'M' : 'MET', 'F' : 'PHE',
        'P' : 'PRO', 'S' : 'SER', 'T' : 'THR', 'W' : 'TRP', 'Y' : 'TYR', 'V' : 'VAL',
        'B' : 'ASX', 'U' : 'SEC', 'Z' : 'GLX'
        }

        self.Three2OneDict = {v: k for k, v in self.One2ThreeDict.items()}

        self.psiblast_parameter = {
        0 : { 'wordSize':2, 'gapOpen':9,  'gapExtend':1, 'scoringMatrix':'PAM30' },
        1 : { 'wordSize':3, 'gapOpen':9,  'gapExtend':1, 'scoringMatrix':'PAM30' },
        2 : { 'wordSize':3, 'gapOpen':10, 'gapExtend':1, 'scoringMatrix':'PAM70' },
        3 : { 'wordSize':3, 'gapOpen':10, 'gapExtend':1, 'scoringMatrix':'BLOSUM80'},
        4 : { 'wordSize':3, 'gapOpen':11, 'gapExtend':1, 'scoringMatrix':'BLOSUM62'}
        }


    def get_fasta(self, pdb_dir='pdb', pdbref=None, chain=['A','B'], out_dir='fasta'):
        """Extract the sequence of the chains and writes a fasta query file for each.

        Args:
            chain (list, optional): Name of the chains in the pdbs
            out_dir (str, optional): name pf the output directory where to store the fast queries
        """
        out_dir = os.path.join(self.work_dir,out_dir)
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)

        if pdbref:
            pdb = os.path.join(self.work_dir, pdb_dir, pdbref)
        else:
            pdbs = os.listdir(os.path.join(self.work_dir, pdb_dir))
            pdbs = list(filter(lambda x: x.endswith('.pdb'), pdbs))
            pdb = os.path.join(self.work_dir, pdb_dir, pdbs[0])

        sqldb = pdb2sql(pdb)
        for c in chain:

            # get the unique residues
            res = sqldb.get_residues(chainID=c)

            # get the one letter resiude
            seq = ''
            count = 0
            for r in res:
                seq += self.Three2OneDict[r[1]]
                count += 1
                if count == 79:
                    seq += '\n'
                    count = 0

            # write the file
            caseID = re.split('_|\.', os.path.basename(pdb))[0]
            fname = os.path.join(out_dir, caseID + '.%s' %c + '.fasta')
            f = open(fname,'w')
            f.write('>%s' %caseID + '.%s\n' %c)
            f.write(seq)
            f.close()
        sqldb.close()


    def configure(self, blast_exe=None, database=None,
        num_threads = 4, evalue=0.0001, comp_based_stats='T',
        max_target_seqs=2000, num_iterations=3, outfmt=7,
        save_each_pssm=True, save_pssm_after_last_round=True):
        """Configure the blast executable, database and psiblast parameters.

        Notes:
            For more details about psiblast paramters, check 'psiblast -help'.

        Args:
            blast_exe (str): Path to the psiblast executable
            database (str) : Path to the Blast database
            num_threads (int): Number of threads (CPUs) to use in the BLAST search
            evalue (float): Expectation value (E) threshold for saving hits
            comp_based_stats (str, int): Use composition-based statistics,
                    0, F or f: no composition-based statistics
                    2, T or t, D or d : Composition-based score adjustment
                    as in Bioinformatics 21:902-911, 2005, conditioned on
                    sequence properties
            max_target_seqs (int): Maximum number of aligned sequences to keep,
                    not applicable for outfmt <= 4
            num_iterations (int): Number of iterations to perform,
                    0 means run until convergence
            save_each_pssm (bool): Save PSSM after each iteration
            save_pssm_after_last_round (bool): Save PSSM after the last database search
        """

        self.blast_exe = blast_exe
        self.blast_config = {
            'db': database,
            'num_threads': num_threads,
            'evalue': evalue,
            'comp_based_stats': comp_based_stats,
            'max_target_seqs': max_target_seqs,
            'num_iterations': num_iterations,
            'save_each_pssm': save_each_pssm,
            'save_pssm_after_last_round': save_pssm_after_last_round,
        }

    def get_pssm(self, fasta_dir='fasta',
                 out_dir='pssm_raw',
                 run=True):

        """Compute the PSSM files

        Args:
            fasta_dir (str, optional): irectory where the fasta queries are stored
            blast (str, optional): path to the psiblast executable
            db (str, optional): path to the blast database
            out_dir (str, optional): output directory where to store the pssm files
            num_iterations (int, optional): number of iterations for the blast calculations
        """

        fasta_dir = os.path.join(self.work_dir,fasta_dir)
        out_dir = os.path.join(self.work_dir,out_dir)
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)

        out_fmt = '7 qseqid qgi qacc qaccver qlen sseqid sallseqid sgi sallgi   ,\
                   sacc saccver sallacc slen qstart qend sstart send qseq sseq  ,\
                   evalue bitscore score length pident nident mismatch positive ,\
                   gapopen gaps ppos frames qframe sframe btop staxids stitle   ,\
                   salltitles sstrand qcovs qcovhsp qcovus'

        for q in os.listdir(fasta_dir):

            # get the fasta quey
            query = os.path.join(fasta_dir,q)
            name = os.path.splitext(os.path.basename(query))[0]

            # set up the output names
            out_ascii_pssm = os.path.join(out_dir,name + '.pssm')
            out_pssm = os.path.join(out_dir,name + '.cptpssm')
            out_xml = os.path.join(out_dir,name + '.xml')

            # get the parameters
            blast_param = self._get_psiblast_parameters(query)

            # set up the psiblast calculation
            psi_cline = NcbipsiblastCommandline(
                               cmd = self.blast_exe,
                               query = query,
                               word_size = blast_param['wordSize'],
                               gapopen = blast_param['gapOpen'],
                               gapextend = blast_param['gapExtend'],
                               matrix = blast_param['scoringMatrix'],
                               outfmt = out_fmt,
                               out_ascii_pssm = out_ascii_pssm,
                               out_pssm = out_pssm,
                               out = out_xml,
                               **self.blast_config
                               )

            # check that it's correct
            psi_cline._validate()

            if run:

                # run the blast query
                psi_cline()

                # copyt the final pssm to its final name
                shutil.copy2(out_ascii_pssm + '.%d' %self.blast_config['num_iterations'], out_ascii_pssm)

                # remove all the other files
                for filename in glob.glob(out_pssm+'.*'):
                    os.remove(filename)
                for filename in glob.glob(out_ascii_pssm+'.*'):
                    os.remove(filename)
                os.remove(out_xml)


    def _get_psiblast_parameters(self,fasta_query):

        f = open(fasta_query)
        data =f.readlines()
        f.close()

        seq = 0
        for l in data[1:]:
            seq += len(l)

        if seq < 30:
            return self.psiblast_parameter[0]
        elif seq < 35:
            return self.psiblast_parameter[1]
        elif seq < 50:
            return self.psiblast_parameter[2]
        elif seq < 85:
            return self.psiblast_parameter[3]
        else:
            return self.psiblast_parameter[4]

    def map_pssm(self, pssm_dir='pssm_raw', pdb_dir='pdb', out_dir='pssm', chain=['A','B']):

        """Map the raw pssm files to the pdb files of the decoys

        Args:
            pssm_dir (str, optional): name pf the directory where the pssm are stored
            out_dir (str, optional): name where thmapped pssm files are stored
            chain (list, optional): name of the chains
        """
        pssm_dir = os.path.join(self.work_dir,pssm_dir)
        out_dir = os.path.join(self.work_dir,out_dir)
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)

        # get list of pssm files
        pf = os.listdir(pssm_dir)
        pssm_files = {}
        for c in chain:
            pssm_files[c] = list(filter(lambda x: x.endswith(c+'.pssm'),pf))[0]

        # get list of pdb files
        pdbs = os.listdir(os.path.join(self.work_dir, pdb_dir))
        pdbs = list(filter(lambda x: x.endswith('.pdb'), pdbs))

        # map pssm and pdb
        for p in pdbs:
            pdb = os.path.join(os.path.join(self.work_dir, pdb_dir), p)
            for c in chain:
                pssm = os.path.join(pssm_dir,pssm_files[c])
                write_mapped_pssm_pdb(pssm, pdb, c, out_dir)

    def get_mapped_pdb(self, pdbpssm_dir='pssm', pdb_dir='pdb',
        pdbnonmatch_dir='pdb_nonmatch'):
        """Write mapped pdb to working folder, by default 'pdb'

        Args:
            pdbpssm_dir (str): directory name where mapped pdb and pssm are stored
            pdbnonmatch_dir (str): directory where backup original pdb files
        """

        pdbpssm_dir = os.path.join(self.work_dir, pdbpssm_dir)
        pdbnonmatch_dir = os.path.join(self.work_dir, pdbnonmatch_dir)

        pdb_files = [f for f in os.listdir(pdbpssm_dir) if f.endswith('.pdb')]

        if pdb_files:
            if not os.path.isdir(pdbnonmatch_dir):
                os.mkdir(pdbnonmatch_dir)

            pdb_dict = {}
            for pdb in pdb_files:
                id, chain, _, _ = pdb.split('.')
                if id not in pdb_dict:
                    pdb_dict[id] = []
                pdb_dict[id].append(chain)

            for id, chains in pdb_dict.items():
                pdb_wd = os.path.join(self.work_dir, pdb_dir, id+".pdb")
                pdb_raw = os.path.join(pdbnonmatch_dir, id+".pdb")
                os.rename(pdb_wd, pdb_raw)
                if len(chains) == 1:
                    pdb_new = os.path.join(pdbpssm_dir, id + "." + chains[0] + '.pssm.pdb')
                    os.rename(pdb_new, pdb_wd)
                else:
                    with open(pdb_wd, 'w') as f:
                        # write REMARK
                        pdb_new = os.path.join(pdbpssm_dir, id + "." + chains[0] + '.pssm.pdb')
                        with open(pdb_new, 'r') as fpdb:
                            lines = [line for line in fpdb if line.startswith('REMARK')]
                            f.writelines(lines)

                        # write each chain
                        for chain in chains:
                            pdb_new = os.path.join(pdbpssm_dir, id + "." + chain + '.pssm.pdb')
                            with open(pdb_new, 'r') as fpdb:
                                lines = [line for line in fpdb
                                    if line.startswith('ATOM') and
                                    line[21:22] == chain]
                                f.writelines(lines)
                            os.remove(pdb_new)
                        f.write('END\n')



#!/usr/bin/env python

"""
Map the the sequence of PDB file to that of PSSM file to get consistent sequence(no gap, no residue X),
and output mapped PSSM file and/or PDB file with the consisitent sequence.

Usage: python map_pssm2pdb.py <input PSSM file> <input PDB file> <ChainID of PDB file> <Output path>
Example: python map_pssm2pdb.py  ./pssm/4CPA.A.pssm  ./pdb/4CPA.pdb  A  ./test

Author: {0} ({1})
"""

import os
import sys
import numpy as np
from Bio import pairwise2

__author__ = "Cunliang Geng"
__email__ = "gengcunliang AT gmail.com"
USAGE = __doc__.format(__author__, __email__)


def check_input(args):
    """Validate user input

    Arguments:
        args {tuple} --  user input arguments
    """
    if len(args) != 4:
        sys.exit(USAGE)


def get_seq_resi_from_pdb(fpdb, chainID):
    """Get FASTA sequence and residue ID from PDB file.

    Arguments:
        fpdb {str} -- input pdb file
        chainID {str} -- the chain to get the sequence and residue number

    Raises:
        ValueError -- ChainID not exist in the pdb file

    Returns:
        list - sequence of the specific chain
        list - residue ID of the specific chain
    """

    res_codes = [
        # 20 canonical amino acids
        ('CYS', 'C'), ('ASP', 'D'), ('SER', 'S'), ('GLN', 'Q'),
        ('LYS', 'K'), ('ILE', 'I'), ('PRO', 'P'), ('THR', 'T'),
        ('PHE', 'F'), ('ASN', 'N'), ('GLY', 'G'), ('HIS', 'H'),
        ('LEU', 'L'), ('ARG', 'R'), ('TRP', 'W'), ('ALA', 'A'),
        ('VAL', 'V'), ('GLU', 'E'), ('TYR', 'Y'), ('MET', 'M'),
        # Non-canonical amino acids
        ('ASX', 'B'), ('SEC', 'U'), ('GLX', 'Z'),
        # ('MSE', 'M'), ('SOC', 'C'),
        # Canonical xNA
        ('  U', 'U'), ('  A', 'A'), ('  G', 'G'), ('  C', 'C'),
        ('  T', 'T'),
    ]

    three_to_one = dict(res_codes)
    _records = set(['ATOM  ', 'HETATM'])

    chainID = chainID.upper()
    sequence = []
    resID = []
    chains = set()
    read = set()
    with open(fpdb, "r") as f:
        for line in f:
            line = line.strip()
            if line[0:6] in _records:
                resn = line[17:20]
                chain = line[21]
                resi = line[22:26]
                icode = line[26]
                r_uid = (resn, chain, resi, icode)
                chains.add(chain)
                if chain == chainID:
                    if r_uid not in read:
                        read.add(r_uid)
                    else:
                        continue
                    aa_resn = three_to_one.get(resn, 'X')
                    sequence.append(aa_resn)
                    resID.append(resi)
        if chainID not in chains:
            raise ValueError(
                "Chain `{}` NOT exist in PDB file '{}'".format(chainID, fpdb))

    return sequence, resID


def get_pssm(fpssm):
    """Get the content of PSSM file.

    Arguments:
        fpssm {str} -- input pssm file

    Raises:
        ValueError -- the line with number of columns not equal to 44

    Returns:
        [2D list] -- pssm
    """
    rule = tuple([str(i) for i in range(10)])
    pssm = []
    with open(fpssm, "r") as f:
        for line in f.readlines():
            line_raw = line
            line = line.strip()
            # only select lines that contain pssm values
            if line.startswith(rule):
                    # TODO parse pssm based on column index
                    # normal PSSM line have 44 columns. Abnormal <44 due to lakcing of gap between numbers.
                if len(line.split()) == 44:
                    pssm.append(line.split())
                else:
                    raise ValueError(
                        "Wrong format of the following line in PSSM file {}:\n{}".format(fpssm, line_raw))

    return pssm


def get_aligned_sequences(seq1, seq2):
    """Align two sequnces using global alignment and return aligned sequences.
        Paramters of global alignment:
            match: 1
            mismtach: 0
            gap open: -2
            gap extend: -1

    Arguments:
        seq1 {str} -- 1st sequence.
        seq2 {str} -- 2nd sequence.

    Returns:
        [numpy array] -- seq1_ali, aligned sequence for seq1
        [numpy array] -- seq2_ali, aligned sequence for seq1
    """

    ali = pairwise2.align.globalxs(seq1, seq2, -2, -1)
    seq1_ali = np.array([i for i in ali[0][0]])
    seq2_ali = np.array([i for i in ali[0][1]])

    return seq1_ali, seq2_ali


def write_pdb_remove_residue(fipdb, fopdb, chainID, resID):
    """Write PDB file with removing some residues.

    Arguments:
        fipdb {str} -- Input PDB file.
        fopdb {str} -- Output PDB file.
        chainID {str} -- The ID of the chain that the to-be-removed residues locates.
        resID {list} -- A list of residue ID to remove.
    """

    fout = open(fopdb, "w")
    resID = [str(i) for i in resID]
    _records = set(['ATOM  ', 'HETATM'])
    with open(fipdb, "r") as f:
        for line in f:
            if line[0:6] in _records:
                chain = line[21]
                resi = line[22:26].strip()
                if chain == chainID and resi in resID:
                    continue
                else:
                    fout.write(line)
            else:
                fout.write(line)
    fout.close()


def write_mapped_pssm_pdb(fpssm, fpdb, chainID, outdir):
    """Map PDB sequence to PSSM sequence to get the consistent sequence,
    and output mapped PSSM and/or PDB file with consistent sequence.

    Arguments:
        fpssm {str} -- input PSSM file
        fpdb {str} -- input PDB file
        chainID {str} -- the specific chain of PDB used to be mapped with PSSM, e.g. A or a
        outdir {str} -- path for output pssm file, e.g. /home/tes

    Output:
        mapped PSSM file:  pdbfilename.chainID.pdb.pssm
        mapped PDB file:   pdbfilename.chainID.pssm.pdb, only output when sequence of input PDB longer than consistent sequence.
    """
    # get pssm and pdb file name
    pdbname = os.path.basename(fpdb)
    pssmname = os.path.basename(fpssm)
    # get pdb sequence and residue numbers
    pdb_seq, pdb_resn = get_seq_resi_from_pdb(fpdb, chainID)
    pdb_seq_str = "".join(pdb_seq)
    # get pssm content and sequnce
    pssm = np.array(get_pssm(fpssm))
    pssm_seq_str = "".join(pssm[:, 1])

    # get aligned seqeuences
    pdb_seq_align, pssm_seq_align = get_aligned_sequences(pdb_seq_str, pssm_seq_str)

    # get indexes for matched and mismatched residues.
    index_match = pdb_seq_align == pssm_seq_align
    index_mismatch = np.logical_not(index_match)

    # make a gap sequence (only "-") and X sequence (only "X") that have same length as pdb/pssm_seq_align
    seqlen = len(pdb_seq_align)
    gap_seq = np.array(["-"] * seqlen)
    resX_seq = np.array(["X"] * seqlen)

    # get index of gap and residue X
    index_gappdb = gap_seq == pdb_seq_align
    index_resXpdb = resX_seq == pdb_seq_align
    index_gappssm = gap_seq == pssm_seq_align
    index_resXpssm = resX_seq == pssm_seq_align
    # get index of normal residues (not gap, not res X) for each sequence
    index_norm_pdb = np.logical_not(np.logical_or(index_gappdb, index_resXpdb))
    index_norm_pssm = np.logical_not(np.logical_or(index_gappssm, index_resXpssm))
    # get index of normal residues for both sequences
    index_norm = np.logical_and(index_norm_pdb, index_norm_pssm)
    # get index of mutated normal residues
    index_mut = np.logical_and(index_mismatch, index_norm)

    # raise warning for mutated normal residues
    if len(set(index_mut)) > 1:
        mut_seq = []
        for i in index_mut:
            if i:
                mut_seq.append("^")
            else:
                mut_seq.append("_")
        try:
            raise Warning("Warning: Mutations exist in following sequences:\n>{pdbname}_{chainid}:\n{pdbseq}\n>{pssmname}:\n{pssmseq}\n{mutseq}\n".format(
                pdbname=pdbname, chainid=chainID, pssmname=pssmname, pdbseq="".join(pdb_seq_align), pssmseq="".join(pssm_seq_align), mutseq="".join(mut_seq)))
        except Warning as e:
            print(e)

    # get pssm with index of normal residues for both sequences, this is the mapped pssm.
    index_norm_nogappssm = index_norm[np.logical_not(index_gappssm)]
    pssm_norm = pssm[index_norm_nogappssm]

    # add the residue number and name of PDB file to the mapped pssm
    # for pssm content, only keep the scoring matrix and information content
    header = ["pdbresi", "pdbresn", "seqresi", "seqresn", "A", "R", "N", "D", "C", "Q", "E", "G", "H", "I", "L", "K", "M", "F", "P", "S", "T", "W", "Y", "V", "IC"]
    header = np.transpose(np.array([[i] for i in header]))
    pdb_resn = [[i] for i in pdb_resn]
    pdb_seq = [[i] for i in pdb_seq]
    index_norm_nogappdb = index_norm[np.logical_not(index_gappdb)]
    resi_pdb = np.array(pdb_resn)[index_norm_nogappdb]
    resn_pdb = np.array(pdb_seq)[index_norm_nogappdb]
    pssm_out = np.concatenate((resi_pdb, resn_pdb, pssm_norm[:, :22], pssm_norm[:, -2:-1]), axis=1)
    pssm_out = np.concatenate((header, pssm_out))

    # write mapped pssm to file which is named with input PDB file name, chain ID and ".pdb.pssm"
    fopssm = os.path.join(outdir, os.path.splitext(pdbname)[0] + "." + chainID.upper() + ".pdb.pssm")
    with open(fopssm, "w") as f:
        for i in pssm_out:
            tmp1 = ["{:>7s}".format(j) for j in i[:4]]
            tmp2 = ["{:>4s}".format(j) for j in i[4:]]
            f.write(" ".join(tmp1+tmp2) + "\n")

    # write mapped PDB file if some residues not exist in the mapped PSSM file
    index_toremove = np.logical_or(np.logical_or(index_gappssm, index_resXpssm), index_resXpdb)
    index_gappdb_resXpssm = np.logical_and(index_gappdb, index_resXpssm)
    index_toremove = np.logical_and(index_toremove, np.logical_not(index_gappdb_resXpssm))
    fopdb = os.path.join(outdir, os.path.splitext(pdbname)[0] + "." + chainID.upper() + ".pssm.pdb")
    if len(np.unique(index_toremove)) == 2:
        # write PDB with removing some resdiues
        index_toremove_nogappdb = index_toremove[np.logical_not(index_gappdb)]
        resn_pdb_remove = np.array(pdb_resn)[index_toremove_nogappdb]
        resn_pdb_remove = [ i.strip() for i in resn_pdb_remove[:,0].tolist() ]
        write_pdb_remove_residue(fpdb, fopdb, chainID, resn_pdb_remove)

        print("Warning: A new PDB file with all chains and consistent sequence of chain {} is generated: {}\n".format(chainID, fopdb))


    
        




if __name__ == "__main__":
    check_input(sys.argv[1:])
    fpssm, fpdb, chainID, outdir = sys.argv[1:]
    # fpssm, fpdb, chainID, outdir = "4CPA.A.pssm", "4CPA.pdb", "A",  "."
    write_mapped_pssm_pdb(fpssm, fpdb, chainID, outdir)



import os
from setuptools import setup

cwd = os.path.abspath(os.path.dirname(__file__))

# To update the package version number, edit pssmgen/__version__.py
version = {}
with open(os.path.join(cwd, 'pssmgen', '__version__.py')) as f:
    exec(f.read(), version)

setup(
    name='PSSMGen',
    description='Generates consistent PSSM and/or PDB files for protein-protein complexes',
    version=version['__version__'],
    author='Nicolas Renaud, Cunliang Geng',
    author_email='n.renaud@esciencecenter.nl, c.geng@esciencecenter.nl',
    url='https://github.com/DeepRank/PSSMGen',
    packages=['pssmgen'],
    license="Apache Software License 2.0",
    keywords='pssmgen',

    install_requires=[
        'numpy >= 1.13',
        'scipy',
        'biopython',
        'pdb2sql'],

    extras_require= {
        'test': ['nose', 'coverage', 'pytest',
                 'pytest-cov','codacy-coverage','coveralls'],
    }
)
